import json
from collections.abc import Callable
from contextlib import contextmanager
from typing import Optional

import urllib3.exceptions

from dasl_api import ApiException


class ConflictError(Exception):
    """
    Simple exception wrapper for 409 errors returned from the API
    """

    def __init__(self, resource: str, identifier: str, message: str) -> None:
        self.resource = resource
        self.identifier = identifier
        self.message = message
        super().__init__(message)

    def __str__(self) -> str:
        return f"Conflict: resource_type='{self.resource}' identifier='{self.identifier}' message='{self.message}'"


class NotFoundError(Exception):
    """
    Simple exception wrapper for 404 errors returned from the API
    """

    def __init__(
        self, identifier: str, message: str, resource_type: str = None
    ) -> None:
        self.identifier = identifier
        self.message = message
        self.resource_type = resource_type
        super().__init__(message)

    def __str__(self) -> str:
        if self.resource_type:
            return f"NotFound: resource_type='{self.resource_type}' identifier='{self.identifier}' message='{self.message}'"
        return f"NotFound: identifier='{self.identifier}' message='{self.message}'"


class BadRequestError(Exception):
    """
    Simple exception wrapper for 400 errors returned from the API
    """

    def __init__(self, message: str) -> None:
        self.message = message
        super().__init__(message)

    def __str__(self) -> str:
        return f"BadRequest: message='{self.message}'"


class UnauthorizedError(Exception):
    """
    Simple exception wrapper for 401 errors returned from the API
    """

    def __init__(self, message: str) -> None:
        self.message = message
        super().__init__(message)

    def __str__(self) -> str:
        return f"Unauthorized: message='{self.message}'"


class ForbiddenError(Exception):
    """
    Simple exception wrapper for 403 errors returned from the API
    """

    def __init__(self, message: str) -> None:
        self.message = message
        super().__init__(message)

    def __str__(self) -> str:
        return f"Forbidden: message='{self.message}'"


class WorkspaceLookupError(Exception):
    """Internal exception wrapper for workspace lookup errors"""

    def __init__(self, message: str, reason: Exception = None) -> None:
        self.message = message
        self.reason = reason
        super().__init__(message)

    def __str__(self) -> str:
        return f"Workspace lookup error: {self.message}"


class RegionMismatchError(Exception):
    """
    Exception raised when authentication fails due to cross-region communication issues.
    This typically occurs when the client is configured with a DASL host in a different
    cloud region than the Databricks workspace.
    """

    def __init__(
        self,
        workspace_url: str,
        current_host: Optional[str],
        expected_host: Optional[str],
        original_error: Exception = None,
    ) -> None:
        self.workspace_url = workspace_url
        self.current_host = current_host
        self.expected_host = expected_host
        self.original_error = original_error

        if current_host is not None and expected_host is not None:
            message = (
                f"Cannot authenticate workspace '{workspace_url}' using host '{current_host}'. "
                f"The workspace is located in a different cloud region. "
                f"Expected host: '{expected_host}'. "
                f"Please specify the correct region or dasl_host when creating the client, "
                f"or omit these parameters to allow auto-configuration."
            )
        else:
            from dasl_client.regions import Regions

            supported_regions = "\n  ".join(["", *Regions.list()])
            message = (
                f"Cannot authenticate workspace '{workspace_url}'. This workspace may be "
                f"located in an unsupported cloud region. Supported regions: {supported_regions}"
            )
        self.message = message
        super().__init__(message)

    def __str__(self) -> str:
        return f"Region mismatch: {self.message}"


def handle_errors(f: Callable) -> Callable:
    """
    A decorator that handles errors returned from the API.

    :param f: the function that could return an API error
    :return: The output from the callable 'f'. If an Api error was raise,
             re-cast it to a library error before re-raising.
    """

    def error_handler(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        except ApiException as e:
            body = json.loads(e.body)
            if e.status == 400:
                raise BadRequestError(body["message"])
            if e.status == 401:
                raise BadRequestError(body["message"])
            if e.status == 403:
                raise ForbiddenError(body["message"])
            if e.status == 404:
                raise NotFoundError(
                    body["identifier"], body["message"], body.get("resourceType")
                )
            if e.status == 409:
                raise ConflictError(
                    body["resourceType"], body["identifier"], body["message"]
                )
            else:
                raise e
        except Exception as e:
            raise e

    return error_handler


@contextmanager
def error_handler(**context):
    """
    A context manager that handles errors returned from the API.

    Within the context, if an API error is raised, it is re-cast to a library
    error before re-raising.

    :param context: Optional context including 'workspace_url' and 'host' for
                    better error messages in region mismatch scenarios.
    """

    workspace_url = context.get("workspace_url")
    current_host = context.get("host")

    try:
        yield
    except ApiException as e:
        body = json.loads(e.body)
        if e.status == 400:
            raise BadRequestError(body["message"])
        if e.status == 401:
            raise BadRequestError(body["message"])
        if e.status == 403:
            raise ForbiddenError(body["message"])
        if e.status == 404:
            raise NotFoundError(
                body["identifier"], body["message"], body.get("resourceType")
            )
        if e.status == 409:
            raise ConflictError(
                body["resourceType"], body["identifier"], body["message"]
            )
        else:
            raise e
    except (urllib3.exceptions.SSLError, urllib3.exceptions.MaxRetryError) as e:
        # Check if this is a region mismatch issue
        _raise_if_region_mismatch(e, workspace_url, current_host)

        # Raise the original error, if not
        raise e
    except Exception as e:
        raise e


def _raise_if_region_mismatch(
    e: Exception, workspace_url: Optional[str], current_host: Optional[str]
):
    if workspace_url:
        err = e

        # Unwrap a WorkspaceLookupError first
        if isinstance(err, WorkspaceLookupError):
            err = err.reason

        # Unwrap a MaxRetryError
        if isinstance(err, urllib3.exceptions.MaxRetryError):
            err = err.reason

        # Check if the error is SSL-related
        is_ssl_error = isinstance(err, urllib3.exceptions.SSLError)

        if is_ssl_error:
            try:
                from dasl_client.metadata import WorkspaceMetadata

                # We call into get_workspace_metadata, NOT get_endpoint_for_workspace because
                # we would risk creating a loop otherwise
                expected_host = WorkspaceMetadata.get_workspace_metadata(workspace_url)
                expected_host = expected_host and expected_host.api_url
                if expected_host and expected_host != current_host:
                    raise RegionMismatchError(
                        workspace_url, current_host, expected_host, e
                    )
            except RegionMismatchError:
                raise
            except Exception:
                # If we can't determine the expected host, but there was no provided host, this
                # region may not be supported
                if current_host is None:
                    raise RegionMismatchError(workspace_url, current_host, None, e)

                # Otherwise, let the caller re-raise original error
                pass



import abc
import base64
import time
from datetime import datetime

from dasl_api import (
    api,
    WorkspaceV1AuthenticateRequest,
    ApiClient,
    WorkspaceV1RequestSecretRequest,
)
from databricks.sdk.errors import ResourceDoesNotExist

from dasl_client.conn.conn import get_base_conn
from dasl_client.errors.errors import error_handler

from databricks.sdk import WorkspaceClient
from typing import Optional

# The minimum age of a conn that we allow. Generally, clients are issued for
# a small batch of operations (one or two requests), not long-running
# operations. This threshold should be tuned to the longest expected operation
# using a single conn.
EXPIRY_OVERLAP_SECONDS = 600


class Authorization(abc.ABC):
    """
    A common interface for Authentication
    """

    @abc.abstractmethod
    def client(self) -> ApiClient:
        raise NotImplementedError("conn method must be implemented")

    def workspace(self) -> str:
        raise NotImplementedError("client method must be implemented")


class ServiceAccountKeyAuth(Authorization):
    """
    Authorisation implementation for Service Account Keys
    """

    def __init__(
        self, workspace: str, service_account_key: str, host: Optional[str] = None
    ):
        self._workspace = workspace
        self._service_account_key = service_account_key
        self._client = get_base_conn(host=host)
        self.expiry: int = int(datetime.now().timestamp())

    def client(self) -> ApiClient:
        """
        Return an API conn that can be used to issue an API request to the
        configured host. The associated bearer token is valid for at least
        EXPIRY_OVERLAP_SECONDS.
        :return: An API conn with valid auth
        """
        if int(datetime.now().timestamp()) > self.expiry - EXPIRY_OVERLAP_SECONDS:
            self.refresh()
        return self._client

    def workspace(self) -> str:
        """
        Return the client associated with this Service Account Key
        :return: The client name.
        """
        return self._workspace

    def refresh(self):
        """
        A helper function to refresh the bearer token used for authentication.
        :return:
        """
        workspace_url = f"https://{self._workspace}"
        host = self._client.configuration.host

        with error_handler(workspace_url=workspace_url, host=host):
            req = WorkspaceV1AuthenticateRequest(
                service_account_key=self._service_account_key
            )
            handler = api.WorkspaceV1Api(api_client=self._client)

            resp = handler.workspace_v1_authenticate(
                workspace=self._workspace, workspace_v1_authenticate_request=req
            )
            self._client.set_default_header("Authorization", f"Bearer {resp.token}")
            verification = api.DbuiV1Api(self._client).dbui_v1_verify_auth()
            self.expiry = verification.expiry


class DatabricksTokenAuth(Authorization):
    """
    Authorization implementation using Databricks Tokens. Note that this will not work on most databricks
    workspaces using Premium or Enterprise tier databricks.
    """

    def __init__(self, workspace: str, token: str, host: Optional[str] = None):
        self._workspace = workspace
        self._databricks_token = token
        self._client = get_base_conn(host=host)
        self.expiry: int = int(datetime.now().timestamp())

    def client(self) -> ApiClient:
        """
        Return an API conn that can be used to issue an API request to the
        configured host. The associated bearer token is valid for at least
        EXPIRY_OVERLAP_SECONDS.
        :return: An API conn with valid auth
        """
        if int(datetime.now().timestamp()) > self.expiry - EXPIRY_OVERLAP_SECONDS:
            self.refresh()
        return self._client

    def workspace(self) -> str:
        """
        Return the client associated with this Databricks Token
        :return: The client name.
        """
        return self._workspace

    def refresh(self):
        """
        A helper function to refresh the bearer token used for authentication.
        :return:
        """
        workspace_url = f"https://{self._workspace}"
        host = self._client.configuration.host

        with error_handler(workspace_url=workspace_url, host=host):
            req = WorkspaceV1AuthenticateRequest(
                databricks_api_token=self._databricks_token
            )
            handler = api.WorkspaceV1Api(api_client=self._client)

            resp = handler.workspace_v1_authenticate(
                workspace=self._workspace, workspace_v1_authenticate_request=req
            )
            self._client.set_default_header("Authorization", f"Bearer {resp.token}")
            verification = api.DbuiV1Api(self._client).dbui_v1_verify_auth()
            self.expiry = verification.expiry


class DatabricksSecretAuth(Authorization):
    """
    Authorization implementation using Databricks Secrets
    """

    def __init__(self, workspace: str, host: Optional[str] = None):
        self._workspace = workspace
        self._client = get_base_conn(host=host)
        self._principal = WorkspaceClient().current_user.me().user_name
        self.expiry: int = int(datetime.now().timestamp())

    def client(self) -> ApiClient:
        """
        Return an API conn that can be used to issue an API request to the
        configured host. The associated bearer token is valid for at least
        EXPIRY_OVERLAP_SECONDS.
        :return: An API conn with valid auth
        """
        if int(datetime.now().timestamp()) > self.expiry - EXPIRY_OVERLAP_SECONDS:
            self.refresh()
        return self._client

    def workspace(self) -> str:
        """
        Return the client associated with this Databricks Token
        :return: The client name.
        """
        return self._workspace

    def refresh(self):
        """
        A helper function to refresh the bearer token used for authentication.
        :return:
        """
        workspace_url = f"https://{self._workspace}"
        host = self._client.configuration.host

        with error_handler(workspace_url=workspace_url, host=host):
            # First we do a pre-authenticate call to refresh the secret. It doesn't really matter if we race
            # here with others, as long as the secret ends up with a recent value in it. The secret can be used
            # more than once, but it does expire
            req = WorkspaceV1RequestSecretRequest(
                principalName=self._principal,
            )
            handler = api.WorkspaceV1Api(api_client=self._client)
            resp = handler.workspace_v1_request_secret(
                workspace=self._workspace, workspace_v1_request_secret_request=req
            )
            secret_name = resp.secret_name
            secret_value = ""
            for tries in range(3):
                try:
                    secret_value = (
                        WorkspaceClient().secrets.get_secret(secret_name, "token").value
                    )
                    break
                except ResourceDoesNotExist:
                    # Maybe there is a race here, let's retry
                    time.sleep(0.5)
            if len(secret_value) == 0:
                raise RuntimeError(f"failed to complete secret auth")

            req = WorkspaceV1AuthenticateRequest(
                databricks_secret=base64.b64decode(secret_value).decode("utf-8"),
            )
            handler = api.WorkspaceV1Api(api_client=self._client)

            resp = handler.workspace_v1_authenticate(
                workspace=self._workspace, workspace_v1_authenticate_request=req
            )
            self._client.set_default_header("Authorization", f"Bearer {resp.token}")
            verification = api.DbuiV1Api(self._client).dbui_v1_verify_auth()
            self.expiry = verification.expiry



import base64
from typing import Optional

from dasl_api import ApiClient, Configuration, WorkspaceV1Api
from dasl_api.models import WorkspaceV1WorkspaceMetadata
from dasl_api.exceptions import ApiException

from .errors.errors import _raise_if_region_mismatch, WorkspaceLookupError


class WorkspaceMetadata:
    """Workspace metadata lookup functionality for auto-detecting API endpoints."""

    @staticmethod
    def get_workspace_metadata(
        workspace_url: str, dasl_host: Optional[str] = None
    ) -> Optional[WorkspaceV1WorkspaceMetadata]:
        """
        Query the workspace metadata endpoint to auto-detect the correct region
        and API endpoint for a given Databricks workspace.

        :param workspace_url: The Databricks workspace URL to lookup
        :param dasl_host: Optional DASL host to use for the lookup. If None, uses default region.
        :returns: WorkspaceV1WorkspaceMetadata if successful, None if workspace not found
        """
        hosts = []
        if dasl_host is None:
            # Use default region for metadata lookup
            from .regions import Regions

            for region in Regions.list():
                hosts.append(Regions.lookup(region))
        else:
            hosts.append(dasl_host)

        last_exception = None
        for host in hosts:
            try:
                metadata = WorkspaceMetadata._get_workspace_metadata(
                    workspace_url, host
                )
                if metadata:
                    return metadata
            except WorkspaceLookupError as e:
                last_exception = e
                continue

        if last_exception:
            raise last_exception
        return None

    @staticmethod
    def _get_workspace_metadata(
        workspace_url: str, dasl_host: str
    ) -> Optional[WorkspaceV1WorkspaceMetadata]:
        try:
            # Create an unauthenticated client for the public metadata endpoint
            configuration = Configuration(host=dasl_host)
            api_client = ApiClient(configuration)
            workspace_api = WorkspaceV1Api(api_client)

            # Base64 encode the workspace URL
            encoded_workspace = base64.urlsafe_b64encode(
                workspace_url.encode()
            ).decode()

            # Call the metadata endpoint
            metadata = workspace_api.workspace_v1_get_workspace_metadata(
                databricks_workspace=encoded_workspace
            )

            return metadata

        except ApiException as e:
            if e.status == 404:
                # Workspace not found or not in supported region
                return None
            elif e.status == 400:
                # Invalid workspace URL
                raise ValueError(f"Invalid workspace URL: {workspace_url}")
            else:
                # Other API errors
                raise WorkspaceLookupError(
                    f"Failed to get workspace metadata: {e}", reason=e
                )
        except Exception as e:
            # Network errors, encoding errors, etc.
            raise WorkspaceLookupError(
                f"Failed to get workspace metadata: {e}", reason=e
            )

    @staticmethod
    def get_endpoint_for_workspace(workspace_url: str) -> Optional[str]:
        """
        Get the API endpoint URL for a workspace.

        :param workspace_url: The Databricks workspace URL
        :returns: API endpoint URL if successful, None if workspace not found
        """
        try:
            metadata = WorkspaceMetadata.get_workspace_metadata(workspace_url)
            if metadata is not None:
                return metadata.api_url
        except Exception as e:
            _raise_if_region_mismatch(e, workspace_url, None)
        return None
